#INDEX\
proc iwidgets::Tabset {} {}
#
# Tabset Widget and the Tab Class
# ----------------------------------------------------------------------
# A Tabset is a widget that contains a set of Tab buttons. 
# It displays these tabs in a row or column depending on it tabpos. 
# When a tab is clicked on, it becomes the only tab in the tab set that 
# is selected. All other tabs are deselected. The Tcl command prefix 
# associated with this tab (through the command tab configure option) 
# is invoked with the tab index number appended to its argument list. 
# This allows the Tabset to control another widget such as a Notebook.
#
# A Tab class is an [incr Tcl] class that displays either an image, 
# bitmap, or label in a graphic object on a canvas. This graphic object 
# can have a wide variety of appearances depending on the options set. 
#
# WISH LIST:
#   This section lists possible future enhancements.
#
#   1) When too many tabs appear, a small scrollbar should appear to
#      move the tabs over.
#
# ----------------------------------------------------------------------
#  AUTHOR: Bill W. Scott                 EMAIL: bscott@spd.dsccc.com
#
#  @(#) $Id: tabset.itk,v 1.2 1996/04/30 15:08:17 mulferts Exp $
# ----------------------------------------------------------------------
#            Copyright (c) 1995 DSC Technologies Corporation
# ======================================================================
# Permission to use, copy, modify, distribute and license this software 
# and its documentation for any purpose, and without fee or written 
# agreement with DSC, is hereby granted, provided that the above copyright 
# notice appears in all copies and that both the copyright notice and 
# warranty disclaimer below appear in supporting documentation, and that 
# the names of DSC Technologies Corporation or DSC Communications 
# Corporation not be used in advertising or publicity pertaining to the 
# software without specific, written prior permission.
# 
# DSC DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING 
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, AND NON-
# INFRINGEMENT. THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND THE
# AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, 
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. IN NO EVENT SHALL 
# DSC BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
# ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
# SOFTWARE.
# ======================================================================

#
# Default resources.
#
option add *Tabset.width          0            widgetDefault
option add *Tabset.height         0            widgetDefault
option add *Tabset.equalTabs      true         widgetDefault
option add *Tabset.tabPos         s            widgetDefault
option add *Tabset.raiseSelect    false        widgetDefault
option add *Tabset.start          4            widgetDefault
option add *Tabset.margin         5            widgetDefault
option add *Tabset.tabBorders     true         widgetDefault
option add *Tabset.bevelAmount    0            widgetDefault
option add *Tabset.padX           4            widgetDefault
option add *Tabset.padY           4            widgetDefault
option add *Tabset.gap            overlap      widgetDefault
option add *Tabset.angle          20           widgetDefault
option add *Tabset.font           fixed        widgetDefault
option add *Tabset.state          normal       widgetDefault
option add *Tabset.disabledForeground #a3a3a3  widgetDefault
option add *Tabset.foreground     black        widgetDefault
option add *Tabset.background     #d9d9d9      widgetDefault
option add *Tabset.selectForeground black      widgetDefault
option add *Tabset.selectBackground #ececec    widgetDefault

#
# Usual options.
#
tk++::usual Tabset {
    keep -backdrop -background -cursor -disabledforeground -font -foreground \
	 -selectbackground -selectforeground
}

# ------------------------------------------------------------------
#                              TABSET
# ------------------------------------------------------------------
class iwidgets::Tabset {
    inherit tk++::Widget
    
    constructor {args} {} 
    destructor {}
    
    itk_option define -width width Width  0 
    itk_option define -equaltabs equalTabs EqualTabs true 
    itk_option define -height height Height  0
    itk_option define -tabpos tabPos TabPos  s
    itk_option define -raiseselect raiseSelect RaiseSelect false
    itk_option define -start start Start 4
    itk_option define -margin margin Margin 5
    itk_option define -tabborders tabBorders TabBorders true
    itk_option define -bevelamount bevelAmount BevelAmount 0
    itk_option define -padx padX PadX 4
    itk_option define -pady padY PadY 4
    itk_option define -gap gap Gap overlap
    itk_option define -angle angle Angle 20
    itk_option define -font font Font fixed
    itk_option define -state state State normal
    itk_option define \
	    -disabledforeground disabledForeground DisabledForeground #a3a3a3
    itk_option define -foreground foreground Foreground black
    itk_option define -background background Background #d9d9d9 
    itk_option define -selectforeground selectForeground Background black
    itk_option define -backdrop backdrop Backdrop white 
    itk_option define \
	    -selectbackground selectBackground Foreground #ececec
    itk_option define -command command Command {} 
    
    public method configure {args} 
    public method add {args} 
    public method delete {args} 
    public method index {index} 
    public method insert {index args} 
    public method prev {} 
    public method next {} 
    public method select {index} 
    public method tabcget {index args} 
    public method tabconfigure {index args} 
    
    protected method _selectName {tabName} 
    
    private method _createTab {args} 
    private method _deleteTabs {fromTab toTab} 
    private method _index {pathList index select} 
    private method _tabConfigure {args} 
    private method _relayoutTabs {} 
    private method _drawBevelBorder {} 
    private method _calcNextTabOffset {tabName} 
    private method _tabBounds {} 
    private method _recalcCanvasGeom {} 
    private method _canvasReconfigure {width height} 
    private method _startMove {x y} 
    private method _moveTabs {x y} 
    private method _endMove {x y} 
    private method _configRelayout {} 
    
    private variable _width 0          ;# Width of the canvas in screen units
    private variable _height 0         ;# Height of the canvas in screen units
    private variable _selectedTop 0    ;# top edge of tab + a margin
    private variable _deselectedTop 0  ;# top edge of tab + a margin&raiseamt
    private variable _selectedLeft 0   ;# left edge of tab + a margin
    private variable _deselectedLeft 0 ;# left edge of tab + a margin&raiseamt
    private variable _tabs {}          ;# our internal list of tabs
    private variable _currTab -1       ;# numerical index # of selected tab
    private variable _uniqueID 0       ;# used to create unique names
    private variable _cmdStr  {}       ;# holds value of itk_option(-command)
    ;# do not know why I need this!
    private variable _canvasWidth 0    ;# set by canvasReconfigure, is can wid
    private variable _canvasHeight 0   ;# set by canvasReconfigure, is can hgt
    
    private variable _anchorX 0        ;# used by mouse scrolling methods
    private variable _anchorY 0        ;# used by mouse scrolling methods
    
    private variable _margin 0         ;# -margin in screen units
    private variable _start  0         ;# -start in screen units
    private variable _gap overlap      ;# -gap in screen units
    
    private variable _relayout false   ;# flag tripped to tell whether to
                                       ;# relayout tabs after the configure
    private variable _skipRelayout false ;# flag that tells whether to skip
                                       ;# relayouting out the tabs. used by
                                       ;# _endMove.
}

#
# Provide a lowercase access method for the Tabset class
#
proc ::iwidgets::tabset {pathName args} {
    uplevel ::iwidgets::Tabset $pathName $args
}

# ----------------------------------------------------------------------
#                              CONSTRUCTOR
# ----------------------------------------------------------------------
body iwidgets::Tabset::constructor {args} {
    #
    # Create the canvas that holds the tabs
    #
    itk_component add canvas {
	canvas $itk_interior.canvas -highlightthickness 0
    } {
	keep -cursor -width -height
    }
    pack $itk_component(canvas) -fill both -expand yes -anchor nw
    
    # ... This gives us a chance to redraw our bevel borders, etc when
    # the size of our canvas changes...
    bind $itk_component(canvas) <Configure> \
	    [code $this _canvasReconfigure %w %h]
    
    # ... Allow button 2 scrolling as in label widget.
    bind $itk_component(canvas) <2>               \
	    [code $this _startMove %x %y]
    bind $itk_component(canvas) <B2-Motion>       \
	    [code $this _moveTabs %x %y]
    bind $itk_component(canvas) <ButtonRelease-2> \
	    [code $this _endMove %x %y]
    
    # @@@ 
    # @@@ Is there a better way?
    # @@@
    bind $itk_component(hull) <Any-Enter> "focus $itk_component(hull)"
    bind $itk_component(hull) <Tab> [code $this next]
    bind $itk_component(hull) <Shift-Tab> [code $this prev]
    
    eval itk_initialize $args
    
    _configRelayout 
    
    _recalcCanvasGeom
    
}

body iwidgets::Tabset::destructor {} {
    foreach tab $_(_tabs) {
	::delete object $tab
    }
}

# ----------------------------------------------------------------------
#                              OPTIONS
# ----------------------------------------------------------------------

# ----------------------------------------------------------------------
# OPTION -width
#
# Sets the width explicitly for the canvas of the tabset
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::width {
    if {$itk_option(-width) != {}} {
    }
    set _(_width) [winfo pixels $itk_interior $itk_option(-width)]
}

# ----------------------------------------------------------------------
# OPTION -equaltabs
#
# If set to true, causes horizontal tabs to be equal in
# in width and vertical tabs to equal in height.
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::equaltabs {
    if {$itk_option(-equaltabs) != {}} {
	set _(_relayout) true
    }
}

# ----------------------------------------------------------------------
# OPTION -height
#
# Sets the height explicitly for the canvas of the tabset
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::height {
    set _(_height) [winfo pixels $itk_interior $itk_option(-height)]
}

# ----------------------------------------------------------------------
# OPTION -tabpos
#
# Sets the tab position of tabs, n, s, e, w
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::tabpos {
    if {$itk_option(-tabpos) != {}} {
	switch $itk_option(-tabpos) {
	    n {
		_tabConfigure -invert true -orient horizontal
	    }
	    s {
		_tabConfigure -invert false -orient horizontal
	    }
	    w {
		_tabConfigure -invert false -orient vertical
	    }
	    e {
		_tabConfigure -invert true -orient vertical
	    }
	    default {
		error "bad anchor position\
			\"$itk_option(-tabpos)\" must be n, s, e, or w"
	    }
	}
    }
}

# ----------------------------------------------------------------------
# OPTION -raiseselect
#
# Sets whether to raise selected tabs slightly
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::raiseselect {
    if {$itk_option(-raiseselect) != {}} {
	set _(_relayout) true
    }
}

# ----------------------------------------------------------------------
# OPTION -start
#
# Sets the offset to start of tab set
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::start {
    if {$itk_option(-start) != {}} {
	set _(_start) [winfo pixels $itk_interior $itk_option(-start)]
	set _(_relayout) true
    } else {
	set _(_start) 4
    }
}

# ----------------------------------------------------------------------
# OPTION -margin
#
# Sets the margin used above n tabs, below s tabs, left of e
# tabs, right of w tabs
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::margin {
    if {$itk_option(-margin) != {}} {
	set _(_margin) [winfo pixels $itk_interior $itk_option(-margin)]
	set _(_relayout) true
    } else {
	set _(_margin) 5
    }
}

# ----------------------------------------------------------------------
# OPTION -tabborders
#
# Boolean that specifies whether to draw the borders of
# the unselected tabs (tabs in background)
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::tabborders {
    if {$itk_option(-tabborders) != {}} {
	_tabConfigure -tabborders $itk_option(-tabborders)
    }
}

# ----------------------------------------------------------------------
# OPTION -bevelamount
#
# Specifies pixel size of tab corners. 0 means no corners.
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::bevelamount {
    if {$itk_option(-bevelamount) != {}} {
	_tabConfigure -bevelamount $itk_option(-bevelamount)
    }
}

# ----------------------------------------------------------------------
# OPTION -padx
#
# Sets the padding in each tab to the left and right of label
# I don't convert for fpixels, since Tab does it for me.
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::padx {
    if {$itk_option(-padx) != {}} {
	_tabConfigure -padx $itk_option(-padx)
    }
}

# ----------------------------------------------------------------------
# OPTION -pady
#
# Sets the padding in each tab to the left and right of label
# I don't convert for fpixels, since Tab does it for me.
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::pady {
    if {$itk_option(-pady) != {}} {
	_tabConfigure -pady $itk_option(-pady)
    }
}

# ----------------------------------------------------------------------
# OPTION -gap
#
# Sets the amount of spacing between tabs in pixels
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::gap {
    if {$itk_option(-gap) != {}} {
	if {$itk_option(-gap) != "overlap"} {
	    set _(_gap) [winfo pixels $itk_interior $itk_option(-gap)]
	} else {
	    set _(_gap) overlap
	}
	set _(_relayout) true 
    } else {
	set _(_gap) overlap
    }
}

# ----------------------------------------------------------------------
# OPTION -angle
#
# Sets the angle of the tab's sides
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::angle {
    if {$itk_option(-angle) != {}} {
	_tabConfigure -angle $itk_option(-angle)
    }
}

# ----------------------------------------------------------------------
# OPTION -font
#
# Sets the font of the tab (SELECTED and UNSELECTED)
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::font {
    if {$itk_option(-font) != {}} {
	_tabConfigure -font $itk_option(-font)
    }
}

# ----------------------------------------------------------------------
# OPTION -state
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::state {
    if {$itk_option(-state) != {}} {
	_tabConfigure -state $itk_option(-state)
    }
}

# ----------------------------------------------------------------------
# OPTION -disabledforeground
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::disabledforeground {
    if {$itk_option(-disabledforeground) != {}} {
	_tabConfigure \
		-disabledforeground $itk_option(-disabledforeground)
    }
}

# ----------------------------------------------------------------------
# OPTION -foreground
#
# Sets the foreground label color of UNSELECTED tabs
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::foreground {
    _tabConfigure -foreground $itk_option(-foreground)
}

# ----------------------------------------------------------------------
# OPTION -background
#
# Sets the background color of UNSELECTED tabs
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::background {
    if {$itk_option(-background) != {}} {
	_tabConfigure -background $itk_option(-background)
    } else {
	_tabConfigure -background \
		[$itk_component(canvas) cget -background]
    }
}

# ----------------------------------------------------------------------
# OPTION -selectforeground
#
# Sets the foreground label color of SELECTED tabs
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::selectforeground {
    _tabConfigure -selectforeground $itk_option(-selectforeground)
}

# ----------------------------------------------------------------------
# OPTION -backdrop
#
# Sets the background color of the Tabset backdrop (behind the tabs)
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::backdrop {
    if {$itk_option(-backdrop) != {}} {
	$itk_component(canvas) configure \
		-background $itk_option(-backdrop)
    }
}

# ----------------------------------------------------------------------
# OPTION -selectbackground
#
# Sets the background color of SELECTED tabs
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::selectbackground {
    if {$itk_option(-selectbackground) != {}} {
    } else {
	#set _selectBackground \
		[$itk_component(canvas) cget -background]
    }
    _tabConfigure -selectbackground $itk_option(-selectbackground)
}

# ----------------------------------------------------------------------
# OPTION -command
#
# The command to invoke when a tab is hit.
# ----------------------------------------------------------------------
configbody iwidgets::Tabset::command {
    if {$itk_option(-command) != {}} {
	set _(_cmdStr) $itk_option(-command)
    }
}

# ----------------------------------------------------------------------
# METHOD: add ?option value...?
#
# Creates a tab and appends it to the list of tabs.
# processes tabconfigure for the tab added.
# ----------------------------------------------------------------------
body iwidgets::Tabset::add {args} {
    set tabName [eval _createTab $args]
    lappend _(_tabs) $tabName
    
    _relayoutTabs
    
    return $tabName
}

# ----------------------------------------------------------------------
# METHOD: configure ?option? ?value option value...?
#
# Acts as an addendum to the itk::Widget::configure method.
#
# Checks the _relayout flag to see if after configures are done
# we need to relayout the tabs.
#
# _(_skipRelayout) is set in the MB2 scroll methods, to avoid constant
# relayout of tabs while dragging the mouse.
# ----------------------------------------------------------------------
body iwidgets::Tabset::configure {args} {
    set result [eval tk++::Archetype::configure $args]

    _configRelayout

    return $result
}

body iwidgets::Tabset::_configRelayout {} {
    # then relayout tabs if necessary
    if { $_(_relayout) } {
	if { $_(_skipRelayout) } {
	} else {
	    _relayoutTabs
	}
	set _(_relayout) false
    }
}

# ----------------------------------------------------------------------
# METHOD: delete index1 ?index2?
#
# Deletes a tab or range of tabs from the tabset
# ----------------------------------------------------------------------
body iwidgets::Tabset::delete {args} {
    if { $_(_tabs) == {} } {
	error "can't delete tabs,\
		no tabs in the tabset named $itk_component(hull)"
    }
    
    set len [llength $args]
    switch $len {
	0 {
	    error "wrong # args: should be\
		    \"$itk_component(hull) delete index1 ?index2?\""
	}
	
	1 {
	    set fromTab [index [lindex $args 0]]
	    if { $fromTab == -1 } {
		error "bad value for index1:\
			[lindex $args 0] in call to delete"
	    }
	    set toTab $fromTab
	    _deleteTabs $fromTab $toTab
	}
	
	2 {
	    set fromTab [index [lindex $args 0]]
	    if { $fromTab == -1 } {
		error "bad value for index1:\
			[lindex $args 0] in call to delete"
	    }
	    set toTab [index [lindex $args 1]]
	    
	    if { $toTab == -1 } {
		error "bad value for index2:\
			[lindex $args 1] in call to delete"
	    }
	    _deleteTabs $fromTab $toTab
	}
	
	default {
	    error "wrong # args: should be\
		    \"$itk_component(hull) delete index1 ?index2?\""
	}
    }
}

# ----------------------------------------------------------------------
# METHOD: index index
#
# Given an index identifier returns the numeric index of the tab
# ----------------------------------------------------------------------
body iwidgets::Tabset::index {index} {
    return [_index $_(_tabs) $index $_(_currTab)]
}

# ----------------------------------------------------------------------
# METHOD: insert index ?option value...?
#
# Inserts a tab before a index. The before tab may
# be specified as a label or a tab position.
# ----------------------------------------------------------------------
body iwidgets::Tabset::insert {index args} {
    if { $_(_tabs) == {} } {
	error "no tab to insert before,\
		tabset '$itk_component(hull)' is empty"
    }
    
    # get the tab
    set tab [index $index]
    
    # catch bad value for before tab.
    if { $tab < 0 || $tab >= [llength $_(_tabs)] } {
	error "bad value $tab for index:\
		should be between 0 and [expr [llength $_(_tabs)] - 1]"
    }
    
    # create the new tab and get its name...
    set tabName [eval _createTab $args]
    
    # grab the name of the tab currently selected. (to keep in sync)
    set currTabName [lindex $_(_tabs) $_(_currTab)]
    
    # insert tabName before $tab
    set _(_tabs) [linsert $_(_tabs) $tab $tabName]
    
    # keep the _(_currTab) in sync with the insert.
    set _(_currTab) [lsearch -exact $_(_tabs) $currTabName]
    
    _relayoutTabs
    
    return $tabName
}

# ----------------------------------------------------------------------
# METHOD: prev
#
# Selects the prev tab. Wraps at first back to last tab.
# ----------------------------------------------------------------------
body iwidgets::Tabset::prev {} {
    if { $_(_tabs) == {} } {
	error "can't goto previous tab,\
		no tabs in the tabset: $itk_component(hull)"
    }
    
    # bump to the previous tab and wrap if necessary
    set prev [expr $_(_currTab) - 1]
    if { $prev < 0 } {
	set prev [expr [llength $_(_tabs)] - 1]
    }
    
    select $prev
    
}

# ----------------------------------------------------------------------
# METHOD: next
#
# Selects the next tab. Wraps at last back to first tab.
# ----------------------------------------------------------------------
body iwidgets::Tabset::next {} {
    if { $_(_tabs) == {} } {
	error "can't goto next tab,\
		no tabs in the tabset: $itk_component(hull)"
    }
    
    # bump to the next tab and wrap if necessary
    set next [expr $_(_currTab) + 1]
    if { $next >= [llength $_(_tabs)] } {
	set next 0
    }
    
    select $next
}

# ----------------------------------------------------------------------
# METHOD: select index
#
# Select a tab by index
#
# Lowers the last _(_currTab) if it existed.
# Then raises the new one if it exists.
#
# Returns numeric index of selection, -1 if failed.
# -------------------------------------------------------------
body iwidgets::Tabset::select {index} {
    if { $_(_tabs) == {} } {
	error "can't activate a tab,\
		no tabs in the tabset: $itk_component(hull)"
    }
    
    # if there is not current selection just ignore trying this selection
    if { $index == "select" && $_(_currTab) == -1 } {
	return -1
    }
    
    # is selection request in range ? 
    set reqTab [index $index]
    if { $reqTab == -1 } {
	error "bad value $index for index:\
		should be from 0 to [expr [llength $_(_tabs)] - 1]"
    }
    
    # If already selected then ignore and return...
    if { $reqTab == $_(_currTab) } {
	return $reqTab
    }
    
    # ---- Deselect
    if { $_(_currTab) != -1 } {
	set currTabName [lindex $_(_tabs) $_(_currTab)]
	$currTabName deselect
	
	# handle different orientations...
	if { $itk_option(-tabpos) == "n" || $itk_option(-tabpos) == "s"} {
	    $currTabName configure -top $_(_deselectedTop)
	} else {
	    $currTabName configure -left $_(_deselectedLeft)
	}
    }
    
    # get the stacking order correct...
    foreach tab $_(_tabs) {
	$tab lower
    }
    
    # set this now so that the -command cmd can do an 'index select'
    # to operate on this tab.
    set _(_currTab) $reqTab
    
    # ---- Select
    set reqTabName [lindex $_(_tabs) $reqTab]
    $reqTabName select
    if { $itk_option(-tabpos) == "n" || $itk_option(-tabpos) == "s"} {
	$reqTabName configure -top $_(_selectedTop)
    } else {
	$reqTabName configure -left $_(_selectedLeft)
    }
    
    set _(_currTab) $reqTab
    
    # invoke any user command string, appended with tab index number
    if { $_(_cmdStr) != {} } {
	set newCmd $_(_cmdStr)
	eval [lappend newCmd $reqTab]
    }
    
    return $reqTab
}

# ----------------------------------------------------------------------
# METHOD: tabcget index ?option? 
#
# Returns the value for the option setting of the tab at index $index.
# ----------------------------------------------------------------------
body iwidgets::Tabset::tabcget {index args} {
    return [lindex [eval tabconfigure $index $args] 2]
}

# ----------------------------------------------------------------------
# METHOD: tabconfigure index ?option? ?value option value?
#
# tabconfigure index : returns configuration list
# tabconfigure index -option : returns option values
# tabconfigure index ?option value option value ...? sets options
#   and returns empty string.
#
# Performs configure on a given tab denoted by index.
#
# Index may be a tab number or a pattern matching the label
# associated with a tab.
# ----------------------------------------------------------------------
body iwidgets::Tabset::tabconfigure {index args} {
    # convert index to numeric
    set tab [index $index]
    
    if { $tab == -1 } {
	error "bad index value:\
		$index for $itk_component(hull) tabconfigure"
    }
    
    set tabName [lindex $_(_tabs) $tab]
    
    set len [llength $args]
    switch $len {
	0 {
	    return [eval $tabName configure]
	}
	1 {
	    return [eval $tabName configure $args]
	}
	default {
	    eval $tabName configure $args
	    _relayoutTabs
	    select select
	}
    }
    return ""
}

# ----------------------------------------------------------------------
# PROTECTED METHOD: _selectName
#
# internal method to allow selection by internal tab name 
# rather than index. This is used by the bind methods
# ----------------------------------------------------------------------
body iwidgets::Tabset::_selectName {tabName} {
    # if the tab is disabled, then ignore this selection...
    if { [$tabName cget -state] == "disabled" } {
	return
    }
    
    set tab [lsearch -exact $_(_tabs) $tabName]
    select $tab
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _createTab
#
# Creates a tab, using unique tab naming, propagates background
# and keeps unique id up to date.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_createTab {args} {
    #
    # create an internal name for the tab: tab0, tab1, etc.
    # these are one-up numbers they do not 
    # correspond to the position the tab is located in.
    #
    set tabName $this-tab$_(_uniqueID)
    
    switch $itk_option(-tabpos) {
	n {
	    set invert true
	    set orient horizontal
	    set x 0
	    set y [expr $_(_margin) + 1]
	}
	s {
	    set invert false
	    set orient horizontal
	    set x 0
	    set y 0
	}
	w {
	    set invert false
	    set orient vertical
	    set x 0 
	    set y 0
	}
	e {
	    set invert true
	    set orient vertical
	    set x [expr $_(_margin) + 1]
	    set y 0
	}
	default {
	    error "bad anchor position\
		    \"$itk_option(-tabpos)\" must be n, s, e, or w"
	}
    }
    
    eval iwidgets::Tab $tabName $itk_component(canvas) \
	    -left             $x \
	    -top              $y \
	    -font             [list $itk_option(-font)] \
	    -background       $itk_option(-background) \
	    -foreground       $itk_option(-foreground) \
	    -selectforeground $itk_option(-selectforeground) \
	    -disabledforeground $itk_option(-disabledforeground) \
	    -selectbackground $itk_option(-selectbackground) \
	    -angle            $itk_option(-angle) \
	    -padx             $itk_option(-padx) \
	    -pady             $itk_option(-pady) \
	    -bevelamount      $itk_option(-bevelamount) \
	    -state            $itk_option(-state) \
	    -tabborders       $itk_option(-tabborders) \
	    -invert           $invert \
	    -orient           $orient \
	    $args
    
    $tabName lower
    
    $itk_component(canvas) \
	    bind $tabName <Button-1> [code $this _selectName $tabName]
    
    incr _(_uniqueID)
    
    return $tabName
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _deleteTabs
#
# Deletes tabs from $fromTab to $toTab.
#
# Operates in two passes, destroys all the widgets
# Then removes the pathName from the tab list
#
# Also keeps the current selection in bounds.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_deleteTabs {fromTab toTab} {
    for { set tab $fromTab } { $tab <= $toTab } { incr tab } {
	set tabName [lindex $_(_tabs) $tab]
	
	# unbind Button-1 from this window name
	$itk_component(canvas) bind $tabName <Button-1> {}
	
	# Destroy the Tab class...
	::delete object $tabName 
    }
    
    # physically remove the tab
    set _(_tabs) [lreplace $_(_tabs) $fromTab $toTab]
    
    # If we deleted a selected tab set our selection to none
    if { $_(_currTab) >= $fromTab && $_(_currTab) <= $toTab } {
	set _(_currTab) -1
	_drawBevelBorder
    }
    
    # make sure _(_currTab) stays in sync with new numbering...
    if { $_(_tabs) == {} } {
	# if deleted only remaining tab,
	# reset current tab to undefined
	set _(_currTab) -1
	
	# or if the current tab was the last tab, it needs come back
    } elseif { $_(_currTab) >= [llength $_(_tabs)] } {
	incr _(_currTab) -1
	if { $_(_currTab) < 0 } {
	    # but only to zero
	    set _(_currTab) 0
	}
    }
    
    _relayoutTabs
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _index
#
# pathList : list of path names to search thru if index is a label
# index    : either number, 'select', 'end', or pattern
# select   : current selection
#
# _index takes takes the value $index converts it to
# a numeric identifier. If the value is not already
# an integer it looks it up in the $pathList array.
# If it fails it returns -1
# ----------------------------------------------------------------------
body iwidgets::Tabset::_index {pathList index select} {
    switch $index {
	select {
	    set number $select
	}
	end {
	    set number [expr [llength $pathList] -1]
	}
	default {
	    # is it an number already?
	    if { [regexp {^[0-9]+$} $index] } {
		set number $index
		if { $number < 0 || $number >= [llength $pathList] } {
		    set number -1
		}
		
		# otherwise it is a label
	    } else {
		# look thru the pathList of pathNames and
		# get each label and compare with index.
		# if we get a match then set number to postion in $pathList
		# and break out.
		# otherwise number is still -1
		set i 0
		set number -1
		foreach pathName $pathList {
		    set label [$pathName cget -label]
		    if { $label == $index } {
			set number $i
			break
		    }
		    incr i
		}
	    }
	}
    }
    
    return $number
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _tabConfigure
# ----------------------------------------------------------------------
body iwidgets::Tabset::_tabConfigure {args} {
    foreach tab $_(_tabs) {
	eval $tab configure $args
    }
    
    set _(_relayout) true
    
    if { $_(_tabs) != {} } {
	select select
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _relayoutTabs
# 
# relays out the tabs with correct spacing...
# ----------------------------------------------------------------------
body iwidgets::Tabset::_relayoutTabs {} {
    if { [llength $_(_tabs)] == 0 } {
	return
    }
    
    # get the max width for fixed width tabs...
    set maxWidth 0
    foreach tab $_(_tabs) {
	set width [$tab labelwidth]
	if { $width > $maxWidth } {
	    set maxWidth $width
	}
    }
    
    # get the max height for fixed height tabs...
    set maxHeight 0
    foreach tab $_(_tabs) {
	set height [$tab labelheight]
	if { $height > $maxHeight } {
	    set maxHeight $height
	}
    }
    
    # get curr tab's name
    set currTabName [lindex $_(_tabs) $_(_currTab)]
    
    # Start with our margin offset in pixels...
    set tabStart $_(_start)
    
    if { $itk_option(-raiseselect) } {
	set raiseAmt 2
    } else {
	set raiseAmt 0
    }
    
    #
    # Depending on the tab layout: n, s, e, or w place the tabs
    # according to orientation, raise, margins, etc.
    #
    switch $itk_option(-tabpos) {
	n {
	    set _(_selectedTop) [expr $_(_margin) + 1]
	    set _(_deselectedTop) [expr $_(_selectedTop) + $raiseAmt]
	    
	    if { $itk_option(-equaltabs) } {
		set tabWidth $maxWidth
	    } else {
		set tabWidth 0
	    }
	    
	    foreach tab $_(_tabs) {
		if { $tab == $currTabName } {
		    $tab configure -left $tabStart -top $_(_selectedTop) \
			    -height $maxHeight -width $tabWidth -anchor c
		} else {
		    $tab configure -left $tabStart -top $_(_deselectedTop) \
			    -height $maxHeight -width $tabWidth -anchor c
		}
		set tabStart [expr $tabStart + [_calcNextTabOffset $tab]]
	    }
	    
	}
	s {
	    set _(_selectedTop) 0
	    set _(_deselectedTop) [expr $_(_selectedTop) - $raiseAmt]
	    
	    if { $itk_option(-equaltabs) } {
		set tabWidth $maxWidth
	    } else {
		set tabWidth 0
	    }
	    
	    foreach tab $_(_tabs) {
		if { $tab == $currTabName } {
		    $tab configure -left $tabStart -top $_(_selectedTop) \
			    -height $maxHeight -width $tabWidth -anchor c
		} else {
		    $tab configure -left $tabStart -top $_(_deselectedTop) \
			    -height $maxHeight -width $tabWidth -anchor c
		}
		set tabStart [expr $tabStart + [_calcNextTabOffset $tab]]
	    }
	    
	}
	w {
	    set _(_selectedLeft) [expr $_(_margin) + 1]
	    set _(_deselectedLeft) [expr $_(_selectedLeft) + $raiseAmt]
	    
	    if { $itk_option(-equaltabs) } {
		set tabHeight $maxHeight
	    } else {
		set tabHeight 0
	    }
	    
	    foreach tab $_(_tabs) {
		# selected
		if { $tab == $currTabName } {
		    $tab configure -top $tabStart -left $_(_selectedLeft) \
			    -height $tabHeight -width $maxWidth -anchor e
		    # deselected
		} else {
		    $tab configure -top $tabStart -left $_(_deselectedLeft) \
			    -height $tabHeight -width $maxWidth -anchor e
		}
		set tabStart [expr $tabStart + [_calcNextTabOffset $tab]]
	    }
	    
	}
	e {
	    set _(_selectedLeft) 0
	    set _(_deselectedLeft) [expr $_(_selectedLeft) - $raiseAmt]
	    
	    if { $itk_option(-equaltabs) } {
		set tabHeight $maxHeight
	    } else {
		set tabHeight 0
	    }
	    
	    foreach tab $_(_tabs) {
		# selected
		if { $tab == $currTabName } {
		    $tab configure -top $tabStart -left $_(_selectedLeft) \
			    -height $tabHeight -width $maxWidth -anchor w
		    # deselected
		} else {
		    $tab configure -top $tabStart -left $_(_deselectedLeft) \
			    -height $tabHeight -width $maxWidth -anchor w
		}
		set tabStart [expr $tabStart + [_calcNextTabOffset $tab]]
	    }
	    
	}
	default {
	    error "bad anchor position\
		    \"$itk_option(-tabpos)\" must be n, s, e, or w"
	}
    }
    
    # put border on & calc our new canvas size...
    _drawBevelBorder
    _recalcCanvasGeom
    
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _drawBevelBorder
# 
# draws the bevel border along tab edge (below selected tab)
# ----------------------------------------------------------------------
body iwidgets::Tabset::_drawBevelBorder {} {
    $itk_component(canvas) delete bevelBorder
    
    switch $itk_option(-tabpos) {
	n {
	    $itk_component(canvas) create line \
		    0 [expr $_(_canvasHeight) - 1] \
		    $_(_canvasWidth) [expr $_(_canvasHeight) - 1] \
		    -fill [iwidgets::colors::topShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	    $itk_component(canvas) create line \
		    0 $_(_canvasHeight) \
		    $_(_canvasWidth) $_(_canvasHeight) \
		    -fill [iwidgets::colors::topShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	}
	s {
	    $itk_component(canvas) create line \
		    0 0 \
		    $_(_canvasWidth) 0 \
		    -fill [iwidgets::colors::bottomShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	    $itk_component(canvas) create line \
		    0 1 \
		    $_(_canvasWidth) 1 \
		    -fill black \
		    -tags bevelBorder
	}
	w {
	    $itk_component(canvas) create line \
		    $_(_canvasWidth) 0 \
		    $_(_canvasWidth) [expr $_(_canvasHeight) - 1] \
		    -fill [iwidgets::colors::topShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	    $itk_component(canvas) create line \
		    [expr $_(_canvasWidth) - 1] 0 \
		    [expr $_(_canvasWidth) - 1] [expr $_(_canvasHeight) - 1] \
		    -fill [iwidgets::colors::topShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	    
	}
	e {
	    $itk_component(canvas) create line \
		    0 0 \
		    0 [expr $_(_canvasHeight) - 1] \
		    -fill black \
		    -tags bevelBorder
	    $itk_component(canvas) create line \
		    1 0 \
		    1 [expr $_(_canvasHeight) - 1] \
		    -fill [iwidgets::colors::bottomShadow $itk_option(-selectbackground)] \
		    -tags bevelBorder
	    
	}
    }
    
    $itk_component(canvas) raise bevelBorder
    if { $_(_currTab) != -1 } {
	set currTabName [lindex $_(_tabs) $_(_currTab)]
	$currTabName raise
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _calcNextTabOffset
# 
# given $tabName, determines the offset in pixels to place
# the next tab's start edge at.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_calcNextTabOffset {tabName} {
    if { $_(_gap) == "overlap" } {
	return [$tabName offset]
    } else {
	return [expr [$tabName majordim] + $_(_gap)]
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _tabBounds
# 
# calculates the bounding box that will completely enclose 
# all the tabs.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_tabBounds {} {
    set bbox { 100000 100000 -10000 -10000 }
    foreach tab $_(_tabs) {
	set tabBBox [$tab bbox]
	# if this left is less use it
	if { [lindex $tabBBox 0] < [lindex $bbox 0] } {
	    set bbox [lreplace $bbox 0 0 [lindex $tabBBox 0]]
	}
	# if this top is greater use it
	if { [lindex $tabBBox 1] < [lindex $bbox 1] } {
	    set bbox [lreplace $bbox 1 1 [lindex $tabBBox 1]]
	}
	# if this right is less use it
	if { [lindex $tabBBox 2] > [lindex $bbox 2] } {
	    set bbox [lreplace $bbox 2 2 [lindex $tabBBox 2]]
	}
	# if this bottom is greater use it
	if { [lindex $tabBBox 3] > [lindex $bbox 3] } {
	    set bbox [lreplace $bbox 3 3 [lindex $tabBBox 3]]
	}
	
    }
    return $bbox
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _recalcCanvasGeom
# 
# Based on size of tabs, recalculates the canvas geometry that
# will hold the tabs.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_recalcCanvasGeom {} {
    if { [llength $_(_tabs)] == 0 } {
	return
    }
    
    set bbox [_tabBounds]
    
    set width [lindex [_tabBounds] 2]
    set height [lindex [_tabBounds] 3]
    
    # now we have the dimensions of all the tabs in the canvas.
    
    
    switch $itk_option(-tabpos) {
	n {
	    # height already includes margin
	    $itk_component(canvas) configure \
		    -width $width \
		    -height $height
	}
	s {
	    $itk_component(canvas) configure \
		    -width $width \
		    -height [expr $height + $_(_margin)]
	}
	w {
	    # width already includes margin
	    $itk_component(canvas) configure \
		    -width $width \
		    -height [expr $height + 1]
	}
	e {
	    $itk_component(canvas) configure \
		    -width [expr $width + $_(_margin)] \
		    -height [expr $height + 1]
	}
	default {
	}
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _canvasReconfigure
# 
# Bound to the reconfigure notify event of a canvas, this 
# method resets canvas's correct width (since we are fill x)
# and redraws the beveled edge border.
# will hold the tabs.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_canvasReconfigure {width height} {
    set _(_canvasWidth) $width
    set _(_canvasHeight) $height
    
    if { [llength $_(_tabs)] > 0 } {
	_drawBevelBorder
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _startMove
# 
# This method is bound to the MB2 down in the canvas area of the
# tab set. This starts animated scrolling of the tabs along their
# major axis.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_startMove {x y} {
    if { $itk_option(-tabpos) == "n" || $itk_option(-tabpos) == "s" } {
	set _(_anchorX) $x
    } else {
	set _(_anchorY) $y
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _moveTabs
# 
# This method is bound to the MB2 motion in the canvas area of the
# tab set. This causes the tabset to move with the mouse.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_moveTabs {x y} {
    if { $itk_option(-tabpos) == "n" || $itk_option(-tabpos) == "s" } {
	set startX [expr $_(_start) + $x - $_(_anchorX)]
	foreach tab $_(_tabs) {
	    $tab configure -left $startX 
	    set startX [expr $startX + [_calcNextTabOffset $tab]]
	}
    } else {
	set startY [expr $_(_start) + $y - $_(_anchorY)]
	foreach tab $_(_tabs) {
	    $tab configure -top $startY 
	    set startY [expr $startY + [_calcNextTabOffset $tab]]
	}
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _endMove
# 
# This method is bound to the MB2 release in the canvas area of the
# tab set. This causes the tabset to end moving tabs.
# ----------------------------------------------------------------------
body iwidgets::Tabset::_endMove {x y} {
    if { $itk_option(-tabpos) == "n" || $itk_option(-tabpos) == "s" } {
	set startX [expr $_(_start) + $x - $_(_anchorX)]
	set _(_skipRelayout) true
	configure -start $startX
	set _(_skipRelayout) false
    } else {
	set startY [expr $_(_start) + $y - $_(_anchorY)]
	set _(_skipRelayout) true
	configure -start $startY
	set _(_skipRelayout) false
    }
}


#==============================================================
# CLASS: Tab
#==============================================================

class iwidgets::Tab {
    constructor {args} {}
    
    destructor {}
    
    public variable bevelamount 0 {}
    public variable state normal {}
    public variable height 0 {}
    public variable width 0 {}
    public variable anchor c {}
    public variable left 0 {}
    public variable top 0 {}
    public variable image {} {}
    public variable bitmap {} {}
    public variable label {} {}
    public variable padx 4 {}
    public variable pady 4 {}
    public variable selectbackground "gray70" {}
    public variable selectforeground "black" {}
    public variable disabledforeground "gray" {}
    public variable background "white" {}
    public variable foreground "black" {}
    public variable orient vertical {}
    public variable invert false {}
    public variable angle 20 {}
    public variable font \
	   "-adobe-helvetica-bold-r-normal--34-240-100-100-p-182-iso8859-1" {}
    public variable tabborders true {}
    
    public method configure {args} 
    public method bbox  {} 
    public method deselect {} 
    public method lower {} 
    public method majordim  {} 
    public method minordim  {} 
    public method offset  {} 
    public method raise {} 
    public method select {}
    public method labelheight {}
    public method labelwidth {} 
    
    private method _makeTab {} 
    private method _createLabel {canvas tagList} 
    private method _makeEastTab {canvas} 
    private method _makeWestTab {canvas} 
    private method _makeNorthTab {canvas}
    private method _makeSouthTab {canvas}
    private method _calcLabelDim {labelItem}
    private method _selectNoRaise {}
    private method _deselectNoLower {}
    
    private variable   _selected false
    private variable   _padX 0
    private variable   _padY 0
    
    private variable   _canvas
    
    # these are in pixels
    private variable   _left 0
    private variable   _width 0
    private variable   _height 0
    private variable   _oldLeft 0
    private variable   _top 0
    private variable   _oldTop 0
    
    private variable   _right
    private variable   _bottom
    
    private variable   _offset
    private variable   _majorDim
    private variable   _minorDim
    
    private variable   _darkShadow
    private variable   _lightShadow
    
    # 
    # graphic components that make up a tab
    #
    private variable   _gRegion
    private variable   _gLabel
    private variable   _gLightOutline {}
    private variable   _gBlackOutline {}
    private variable   _gTopLine
    private variable   _gTopLineShadow
    private variable   _gLightShadow
    private variable   _gDarkShadow
    
    private variable   _labelWidth 0
    private variable   _labelHeight 0
    
    private variable   _labelXOrigin 0
    private variable   _labelYOrigin 0
    
    private variable   _just left
    
    private variable   _configTripped true
    
    common _tan
    
    set _tan(0)  0.0
    set _tan(1)  0.0175
    set _tan(2)  0.0349
    set _tan(3)  0.0524
    set _tan(4)  0.0699
    set _tan(5)  0.0875
    set _tan(6)  0.1051
    set _tan(7)  0.1228
    set _tan(8)  0.1405
    set _tan(9)  0.1584
    set _tan(10) 0.1763
    set _tan(11) 0.1944
    set _tan(12) 0.2126
    set _tan(13) 0.2309
    set _tan(14) 0.2493
    set _tan(15) 0.2679
    set _tan(16) 0.2867
    set _tan(17) 0.3057
    set _tan(18) 0.3249
    set _tan(19) 0.3443
    set _tan(20) 0.3640
    set _tan(21) 0.3839
    set _tan(22) 0.4040
    set _tan(23) 0.4245
    set _tan(24) 0.4452
    set _tan(25) 0.4663
    set _tan(26) 0.4877
    set _tan(27) 0.5095
    set _tan(28) 0.5317
    set _tan(29) 0.5543
    set _tan(30) 0.5774
    set _tan(31) 0.6009
    set _tan(32) 0.6294
    set _tan(33) 0.6494
    set _tan(34) 0.6745
    set _tan(35) 0.7002
    set _tan(36) 0.7265
    set _tan(37) 0.7536
    set _tan(38) 0.7813
    set _tan(39) 0.8098
    set _tan(40) 0.8391
    set _tan(41) 0.8693
    set _tan(42) 0.9004
    set _tan(43) 0.9325
    set _tan(44) 0.9657
    set _tan(45) 1.0
}

# ----------------------------------------------------------------------
#                              CONSTRUCTOR
# ----------------------------------------------------------------------
body iwidgets::Tab::constructor {args} {
    
    #import add iwidgets::Tabset
    
    set _(_canvas) [lindex $args 0]
    set args [lrange $args 1 [llength $args]]
    
    set _(_darkShadow)  [iwidgets::colors::bottomShadow $v(-selectbackground)]
    set _(_lightShadow) [iwidgets::colors::topShadow $v(-selectbackground)]
    
    if { $args != "" } {
	eval configure $args
    }
}

# ----------------------------------------------------------------------
#                              DESTRUCTOR
# ----------------------------------------------------------------------
body iwidgets::Tab::destructor {} {
    if { [winfo exists $_(_canvas)] } {
	$_(_canvas) delete $this
    }
}

# ----------------------------------------------------------------------
#                              OPTIONS
# ----------------------------------------------------------------------
#
# Note, we trip _configTripped for every option that requires the tab
# to be remade.
#
# ----------------------------------------------------------------------
# OPTION -bevelamount
#
# Specifies the size of tab corners. A value of 0 with angle set 
# to 0 results in square tabs. A bevelAmount of 4, means that the 
# tab will be drawn with angled corners that cut in 4 pixels from 
# the edge of the tab. The default is 0.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::bevelamount { 
}

# ----------------------------------------------------------------------
# OPTION -state
#
# sets the active state of the tab. specifying normal allows 
# the tab to be selectable. Specifying disabled disables the tab, 
# causing its image, bitmap, or label to be drawn with the 
# disabledForeground color.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::state { 
}

# ----------------------------------------------------------------------
# OPTION -height
#
# the height of the tab. if 0, uses the font label height.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::height {
    set _(_height) [winfo pixels $_(_canvas) $v(-height)]
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -width
#
# The width of the tab. If 0, uses the font label width.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::width {
    set _(_width) [winfo pixels $_(_canvas) $v(-width)]
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -anchor
#
# Where the text in the tab will be anchored: n,nw,ne,s,sw,se,e,w,center
# ----------------------------------------------------------------------
configbody iwidgets::Tab::anchor { 
}

# ----------------------------------------------------------------------
# OPTION -left
#
# Specifies the left edge of the tab's bounding box. This value 
# may have any of the forms acceptable to Tk_GetPixels.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::left {
    
    # get into pixels
    set _(_left) [winfo pixels $_(_canvas) $v(-left)]
    
    # move by offset from last setting
    $_(_canvas) move $this [expr $_(_left) - $_(_oldLeft)] 0
    
    # update old for next time
    set _(_oldLeft) $_(_left)
}

# ----------------------------------------------------------------------
# OPTION -top
#
# Specifies the topedge of the tab's bounding box. This value may 
# have any of the forms acceptable to Tk_GetPixels.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::top {
    
    # get into pixels
    set _(_top) [winfo pixels $_(_canvas) $v(-top)]
    
    # move by offset from last setting
    $_(_canvas) move $this 0 [expr $_(_top) - $_(_oldTop)]
    
    # update old for next time
    set _(_oldTop) $_(_top)
}

# ----------------------------------------------------------------------
# OPTION -image
#
# Specifies the imageto display in the tab. 
# Images are created with the image create command. 
# ----------------------------------------------------------------------
configbody iwidgets::Tab::image { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -bitmap
#
# If bitmap is an empty string, specifies the bitmap to display in 
# the tab. Bitmap may be of any of the forms accepted by Tk_GetBitmap. 
# ----------------------------------------------------------------------
configbody iwidgets::Tab::bitmap { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -label
#
# If image is an empty string and bitmap is an empty string, 
# it specifies a text string to be placed in the tab's label. 
# This label serves as an additional identifier used to reference 
# the tab. Label may be used for the index value in widget commands.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::label { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -padx
#
# Horizontal padding around the label (text, image, or bitmap).
# ----------------------------------------------------------------------
configbody iwidgets::Tab::padx {
    set _(_configTripped) true
    set _(_padX) [winfo pixels $_(_canvas) $v(-padx)]
}

# ----------------------------------------------------------------------
# OPTION -pady
#
# Vertical padding around the label (text, image, or bitmap).
# ----------------------------------------------------------------------
configbody iwidgets::Tab::pady {
    set _(_configTripped) true
    set _(_padY) [winfo pixels $_(_canvas) $v(-pady)]
}

# ----------------------------------------------------------------------
# OPTION -selectbackground
# ----------------------------------------------------------------------
configbody iwidgets::Tab::selectbackground {
    set _(_darkShadow)  [iwidgets::colors::bottomShadow $v(-selectbackground)]
    set _(_lightShadow) [iwidgets::colors::topShadow $v(-selectbackground)]
    
    if { $_(_selected) } {
	_selectNoRaise
    } else {
	_deselectNoLower
    }
}

# ----------------------------------------------------------------------
# OPTION -selectforeground
#
# Foreground of tab when selected
# ----------------------------------------------------------------------
configbody iwidgets::Tab::selectforeground { 
    if { $_(_selected) } {
	_selectNoRaise
    } else {
	_deselectNoLower
    }
}

# ----------------------------------------------------------------------
# OPTION -disabledforeground
#
# Background of tab when -state is disabled
# ----------------------------------------------------------------------
configbody iwidgets::Tab::disabledforeground { 
    if { $_(_selected) } {
	_selectNoRaise
    } else {
	_deselectNoLower
    }
}

# ----------------------------------------------------------------------
# OPTION -background
#
# Normal background of tab.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::background { 
    
    if { $_(_selected) } {
	_selectNoRaise
    } else {
	_deselectNoLower
    }
    
}

# ----------------------------------------------------------------------
# OPTION -foreground
#
# Foreground of tabs when in normal unselected state
# ----------------------------------------------------------------------
configbody iwidgets::Tab::foreground { 
    if { $_(_selected) } {
	_selectNoRaise
    } else {
	_deselectNoLower
    }
}

# ----------------------------------------------------------------------
# OPTION -orient
#
# Specifies the orientation of the tab. Orient can be either 
# horizontal or vertical. 
# ----------------------------------------------------------------------
configbody iwidgets::Tab::orient { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -invert
#
# Specifies the direction to draw the tab. If invert is true, 
# it draws horizontal tabs upside down and vertical tabs opening 
# to the left (pointing right). The value may have any of the 
# forms accepted by the Tcl_GetBoolean, such as true, 
# false, 0, 1, yes, or no.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::invert { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -angle
#
# Specifes the angle of slope from the inner edge to the outer edge 
# of the tab. An angle of 0 specifies square tabs. Valid ranges are 
# 0 to 45 degrees inclusive. Default is 15 degrees. If this option 
# is specified as an empty string (the default), then the angle 
# option for the overall Tabset is used.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::angle {
    if {$v(-angle) < 0 || $v(-angle) > 45 } {
	error "bad angle: must be between 0 and 45"
    }
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# OPTION -font
#
# Font for tab text.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::font { 
}


# ----------------------------------------------------------------------
# OPTION -tabborders
#
# Specifies whether to draw the borders of a deselected tab. 
# Specifying true (the default) draws these borders, 
# specifying false disables this drawing. If the tab is in 
# its selected state this option has no effect. 
# The value may have any of the forms accepted by the 
# Tcl_GetBoolean, such as true, false, 0, 1, yes, or no.
# ----------------------------------------------------------------------
configbody iwidgets::Tab::tabborders { 
    set _(_configTripped) true
}

# ----------------------------------------------------------------------
# METHOD: configure ?option value?
#
# Configures the Tab, checks a configTripped flag to see if the tab
# needs to be remade. We take the easy way since it is so inexpensive
# to delete canvas items and remake them.
# ----------------------------------------------------------------------
body iwidgets::Tab::configure {args} {
    set len [llength $args]
    
    switch $len {
	0 {
	    set result [tcl++::Object::configure]
	    return $result
	}
	1 {
	    set result [eval tcl++::Object::configure $args]
	    return $result
	}
	default {
	    eval tcl++::Object::configure $args
	    if { $_(_configTripped) } {
		_makeTab 
		set _(_configTripped) false
	    }
	    return ""
	}
    }
}

# ----------------------------------------------------------------------
# METHOD: bbox
#
# Returns the bounding box of the tab
# ----------------------------------------------------------------------
body iwidgets::Tab::bbox {} {
    return [lappend bbox $_(_left) $_(_top) $_(_right) $_(_bottom)]
}
# ----------------------------------------------------------------------
# METHOD: deselect
#
# Causes the given tab to be drawn as deselected and lowered
# ----------------------------------------------------------------------
body iwidgets::Tab::deselect {} {
    $_(_canvas) lower $this
    _deselectNoLower
}

# ----------------------------------------------------------------------
# METHOD: lower
#
# Lowers the tab below all others in the canvas.
#
# This is used as our tag name on the canvas.
# ----------------------------------------------------------------------
body iwidgets::Tab::lower {} {
    $_(_canvas) lower $this
}

# ----------------------------------------------------------------------
# METHOD: majordim
#
# Returns the width for horizontal tabs and the height for
# vertical tabs.
# ----------------------------------------------------------------------
body iwidgets::Tab::majordim {} {
    return $_(_majorDim)  
}

# ----------------------------------------------------------------------
# METHOD: minordim
#
# Returns the height for horizontal tabs and the width for
# vertical tabs.
# ----------------------------------------------------------------------
body iwidgets::Tab::minordim {} {
    return $_(_minorDim)  
}

# ----------------------------------------------------------------------
# METHOD: offset
#
# Returns the width less the angle offset. This allows a 
# geometry manager to ask where to place a sibling tab.
# ----------------------------------------------------------------------
body iwidgets::Tab::offset {} {
    return $_(_offset)  
}

# ----------------------------------------------------------------------
# METHOD: raise
#
# Raises the tab above all others in the canvas.
#
# This is used as our tag name on the canvas.
# ----------------------------------------------------------------------
body iwidgets::Tab::raise {} {
    $_(_canvas) raise $this
}

# ----------------------------------------------------------------------
# METHOD: select
#
# Causes the given tab to be drawn as selected. 3d shadows are
# turned on and top line and top line shadow are drawn in sel
# bg color to hide them.
# ----------------------------------------------------------------------
body iwidgets::Tab::select {} {
    $_(_canvas) raise $this
    _selectNoRaise
}

# ----------------------------------------------------------------------
# METHOD: labelheight
#
# Returns the height of the tab's label in its current font.
# ----------------------------------------------------------------------
body iwidgets::Tab::labelheight {} {
    if {$_(_gLabel) != 0} {
	set labelBBox [$_(_canvas) bbox $_(_gLabel)]
	set labelHeight [expr [lindex $labelBBox 3] - [lindex $labelBBox 1]]
    } else {
	set labelHeight 0
    }
    return $labelHeight
}

# ----------------------------------------------------------------------
# METHOD: labelwidth
#
# Returns the width of the tab's label in its current font.
# ----------------------------------------------------------------------
body iwidgets::Tab::labelwidth {} {
    if {$_(_gLabel) != 0} {
	set labelBBox [$_(_canvas) bbox $_(_gLabel)]
	set labelWidth [expr [lindex $labelBBox 2] - [lindex $labelBBox 0]]
    } else {
	set labelWidth 0
    }
    return $labelWidth
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _selectNoRaise
#
# Draws tab as selected without raising it.
# ----------------------------------------------------------------------
body iwidgets::Tab::_selectNoRaise {} {
    if { ! [info exists _(_gRegion)] } {
	return
    }
    
    $_(_canvas) itemconfigure $_(_gRegion) -fill $v(-selectbackground)
    $_(_canvas) itemconfigure $_(_gTopLine) -fill $v(-selectbackground)
    $_(_canvas) itemconfigure $_(_gTopLineShadow) -fill $v(-selectbackground)
    $_(_canvas) itemconfigure $_(_gLightShadow) -fill $_(_lightShadow)
    $_(_canvas) itemconfigure $_(_gDarkShadow) -fill $_(_darkShadow)
    
    if { $_(_gLightOutline) != {} } {
	$_(_canvas) itemconfigure $_(_gLightOutline) -fill $_(_lightShadow)
    }
    if { $_(_gBlackOutline) != {} } {
	$_(_canvas) itemconfigure $_(_gBlackOutline) -fill black
    }
    
    if { $v(-state) == "normal" } {
	if { $v(-image) != {}} {
	    # do nothing for now
	} elseif { $v(-bitmap) != {}} {
	    $_(_canvas) itemconfigure $_(_gLabel) \
		    -foreground $v(-selectforeground) \
		    -background $v(-selectbackground) 
	} else {
	    $_(_canvas) itemconfigure $_(_gLabel) -fill $v(-selectforeground)
	}
    } else {
	if { $v(-image) != {}} {
	    # do nothing for now
	} elseif { $v(-bitmap) != {}} {
	    $_(_canvas) itemconfigure $_(_gLabel) \
		    -foreground $v(-disabledforeground) \
		    -background $v(-selectbackground) 
	} else {
	    $_(_canvas) itemconfigure $_(_gLabel) -fill $v(-disabledforeground)
	}
    }
    
    set _(_selected) true
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _deselectNoLower
#
# Causes the given tab to be drawn as deselected. 3d shadows are
# removed and top line and top line shadow are drawn in visible
# colors to reveal them.
# ----------------------------------------------------------------------
body iwidgets::Tab::_deselectNoLower {} {
    if { ! [info exists _(_gRegion)] } {
	return
    }
    
    $_(_canvas) itemconfigure $_(_gRegion) -fill $v(-background)
    $_(_canvas) itemconfigure $_(_gTopLine) -fill black
    $_(_canvas) itemconfigure $_(_gTopLineShadow) -fill $_(_darkShadow)
    $_(_canvas) itemconfigure $_(_gLightShadow) -fill $v(-background)
    $_(_canvas) itemconfigure $_(_gDarkShadow) -fill $v(-background)
    
    if { $v(-tabborders) } {
	if { $_(_gLightOutline) != {} } {
	    $_(_canvas) itemconfigure $_(_gLightOutline) -fill $_(_lightShadow)
	}
	if { $_(_gBlackOutline) != {} } {
	    $_(_canvas) itemconfigure $_(_gBlackOutline) -fill black
	}
    } else {
	if { $_(_gLightOutline) != {} } {
	    $_(_canvas) itemconfigure $_(_gLightOutline) -fill $v(-background)
	}
	if { $_(_gBlackOutline) != {} } {
	    $_(_canvas) itemconfigure $_(_gBlackOutline) -fill $v(-background)
	}
    }
    
    
    if { $v(-state) == "normal" } {
	if { $v(-image) != {}} {
	    # do nothing for now
	} elseif { $v(-bitmap) != {}} {
	    $_(_canvas) itemconfigure $_(_gLabel) \
		    -foreground $v(-foreground) \
		    -background $v(-background) 
	} else {
	    $_(_canvas) itemconfigure $_(_gLabel) -fill $v(-foreground)
	}
    } else {
	if { $v(-image) != {}} {
	    # do nothing for now
	} elseif { $v(-bitmap) != {}} {
	    $_(_canvas) itemconfigure $_(_gLabel) \
		    -foreground $v(-disabledforeground) \
		    -background $v(-background) 
	} else {
	    $_(_canvas) itemconfigure $_(_gLabel) -fill $v(-disabledforeground)
	}
    }
    
    set _(_selected) false
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _makeTab
# ----------------------------------------------------------------------
body iwidgets::Tab::_makeTab {} {
    if { $v(-orient) == "horizontal" } {
	if { $v(-invert) } {
	    _makeNorthTab $_(_canvas)
	} else {
	    _makeSouthTab $_(_canvas)
	}
    } elseif { $v(-orient) == "vertical" } {
	if { $v(-invert) } {
	    _makeEastTab $_(_canvas)
	} else {
	    _makeWestTab $_(_canvas)
	}
    } else {
	error "bad value for option -orient"
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _createLabel
#
# Creates the label for the tab. Can be either a text label
# or a bitmap label.
# ----------------------------------------------------------------------
body iwidgets::Tab::_createLabel {canvas tagList} {
    if { $v(-image) != {}} {
	set _(_gLabel) [$canvas create image \
		0 0 \
		-image $v(-image) \
		-anchor nw \
		-tags $tagList \
		]
    } elseif { $v(-bitmap) != {}} {
	set _(_gLabel) [$canvas create bitmap \
		0 0 \
		-bitmap $v(-bitmap) \
		-anchor nw \
		-tags $tagList \
		]
    } else {
	set _(_gLabel) [$canvas create text \
		0 0 \
		-text $v(-label) \
		-font $v(-font) \
		-anchor nw \
		-tags $tagList \
		]
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _makeEastTab
#
# Makes a tab that hangs to the east and opens to the west.
# ----------------------------------------------------------------------
body iwidgets::Tab::_makeEastTab {canvas} {
    ::variable _tan
    $canvas delete $this
    set _(_gLightOutline) {}
    set _(_gBlackOutline) {}
    
    lappend tagList $this TAB
    
    _createLabel $canvas $tagList
    
    _calcLabelDim $_(_gLabel)
    
    
    set right  [expr $_(_left) + $_(_labelWidth)]
    # now have _left, _top, right...
    
    # Turn off calculating angle tabs on Vertical orientations
    #set angleOffset [expr $_(_labelHeight) * $_tan($v(-angle))]
    set angleOffset 0
    
    set outerTop $_(_top)
    set outerBottom \
	    [expr $outerTop + $angleOffset + $_(_labelHeight) + $angleOffset]
    set innerTop [expr $outerTop + $angleOffset]
    set innerBottom [expr $outerTop + $angleOffset + $_(_labelHeight)]
    
    # now have _left, _top, right, outerTop, innerTop,
    # innerBottom, outerBottom, width, height
    
    set bottom $innerBottom
    # tab area... gets filled either white or selected
    # done
    set _(_gRegion) [$canvas create polygon \
	    $_(_left) $outerTop \
	    [expr $right - $v(-bevelamount)] $innerTop \
	    $right [expr $innerTop + $v(-bevelamount)] \
	    $right [expr $innerBottom - $v(-bevelamount)] \
	    [expr $right - $v(-bevelamount)] $innerBottom \
	    $_(_left) $outerBottom \
	    $_(_left) $outerTop \
	    -tags $tagList  \
	    ]
    
    # lighter shadow (left edge)
    set _(_gLightShadow) [$canvas create line \
	    [expr $_(_left) - 3] [expr $outerTop + 1] \
	    [expr $right - $v(-bevelamount)] [expr $innerTop + 1] \
	    -tags $tagList \
	    ]
    
    # darker shadow (bottom and right edges)
    set _(_gDarkShadow) [$canvas create line \
	    [expr $right - $v(-bevelamount)] [expr $innerTop + 1] \
	    [expr $right - 1] [expr $innerTop + $v(-bevelamount)] \
	    [expr $right - 1] [expr $innerBottom - $v(-bevelamount)] \
	    [expr $right - $v(-bevelamount)] [expr $innerBottom - 1] \
	    [expr $_(_left) - 3] [expr $outerBottom - 1] \
	    -tags $tagList \
	    ]
    
    # outline of tab
    set _(_gLightOutline) [$canvas create line \
	    $_(_left) $outerTop \
	    [expr $right - $v(-bevelamount)] $innerTop \
	    -tags $tagList \
	    ]
    # outline of tab
    set _(_gBlackOutline) [$canvas create line \
	    [expr $right - $v(-bevelamount)] $innerTop \
	    $right [expr $innerTop + $v(-bevelamount)] \
	    $right [expr $innerBottom - $v(-bevelamount)] \
	    [expr $right - $v(-bevelamount)] $innerBottom \
	    $_(_left) $outerBottom \
	    $_(_left) $outerTop \
	    -tags $tagList \
	    ]
    
    # line closest to the edge
    set _(_gTopLineShadow) [$canvas create line \
	    $_(_left) $outerTop \
	    $_(_left) $outerBottom \
	    -tags $tagList \
	    ]
    
    # next line down
    set _(_gTopLine) [$canvas create line \
	    [expr $_(_left) + 1] [expr $outerTop + 2] \
	    [expr $_(_left) + 1] [expr $outerBottom - 1] \
	    -tags $tagList  \
	    ]
    
    $canvas coords $_(_gLabel) [expr $_(_left) + $_(_labelXOrigin)] \
	    [expr $innerTop + $_(_labelYOrigin)]
    
    if { $v(-image) != {} || $v(-bitmap) != {} } {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor)
    } else {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) -justify $_(_just)
    }
    
    $canvas raise $_(_gLabel) $_(_gRegion)
    
    
    set _(_offset)    [expr $innerBottom - $outerTop]
    # height
    set _(_majorDim)  [expr $outerBottom - $outerTop]
    # width
    set _(_minorDim) [expr $right - $_(_left)]
    
    set _(_right)   $right 
    set _(_bottom)  $outerBottom
    
    # draw in correct state...
    if { $_(_selected) } {
	select
    } else {
	deselect
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _makeWestTab
#
# Makes a tab that hangs to the west and opens to the east.
# ----------------------------------------------------------------------
body iwidgets::Tab::_makeWestTab {canvas} {
    ::variable _tan
    $canvas delete $this
    set _(_gLightOutline) {}
    set _(_gBlackOutline) {}
    
    lappend tagList $this TAB
    
    _createLabel $canvas $tagList
    _calcLabelDim $_(_gLabel)
    
    set right  [expr $_(_left) + $_(_labelWidth)]
    # now have _left, _top, right...
    
    # Turn off calculating angle tabs on Vertical orientations
    #set angleOffset [expr $_(_labelHeight) * $_tan($v(-angle))]
    set angleOffset 0
    
    set outerTop $_(_top)
    set outerBottom \
	    [expr $outerTop + $angleOffset + $_(_labelHeight) + $angleOffset]
    set innerTop [expr $outerTop + $angleOffset]
    set innerBottom [expr $outerTop + $angleOffset + $_(_labelHeight)]
    
    # now have _left, _top, right, outerTop, innerTop,
    # innerBottom, outerBottom, width, height
    
    # tab area... gets filled either white or selected
    # done
    set _(_gRegion) [$canvas create polygon \
	    $right $outerTop \
	    [expr $_(_left) + $v(-bevelamount)] $innerTop \
	    $_(_left) [expr $innerTop + $v(-bevelamount)] \
	    $_(_left) [expr $innerBottom - $v(-bevelamount)]\
	    [expr $_(_left) + $v(-bevelamount)] $innerBottom \
	    $right $outerBottom \
	    $right $outerTop \
	    -tags $tagList  \
	    ]
    # lighter shadow (left edge)
    set _(_gLightShadow) [$canvas create line \
	    $right [expr $outerTop+1] \
	    [expr $_(_left) + $v(-bevelamount)] [expr $innerTop + 1] \
	    [expr $_(_left) + 1] [expr $innerTop + $v(-bevelamount)] \
	    [expr $_(_left) + 1] [expr $innerBottom - $v(-bevelamount)] \
	    -tags $tagList \
	    ]
    
    # darker shadow (bottom and right edges)
    set _(_gDarkShadow) [$canvas create line \
	    [expr $_(_left) + 1] [expr $innerBottom - $v(-bevelamount)] \
	    [expr $_(_left) + $v(-bevelamount)] [expr $innerBottom - 1] \
	    $right [expr $outerBottom - 1] \
	    -tags $tagList \
	    ]
    
    # outline of tab -- lighter top left sides
    set _(_gLightOutline) [$canvas create line \
	    $right $outerTop \
	    [expr $_(_left) + $v(-bevelamount)] $innerTop \
	    $_(_left) [expr $innerTop + $v(-bevelamount)] \
	    $_(_left) [expr $innerBottom - $v(-bevelamount)]\
	    -tags $tagList \
	    ]
    # outline of tab -- darker bottom side
    set _(_gBlackOutline) [$canvas create line \
	    $_(_left) [expr $innerBottom - $v(-bevelamount)]\
	    [expr $_(_left) + $v(-bevelamount)] $innerBottom \
	    $right $outerBottom \
	    $right $outerTop \
	    -tags $tagList \
	    ]
    
    # top of tab
    set _(_gTopLine) [$canvas create line \
	    [expr $right + 1] $outerTop \
	    [expr $right + 1] $outerBottom \
	    -tags $tagList  \
	    ]
    
    # line below top of tab
    set _(_gTopLineShadow) [$canvas create line \
	    $right $outerTop \
	    $right $outerBottom \
	    -tags $tagList \
	    ]
    
    $canvas coords $_(_gLabel) [expr $_(_left) + $_(_labelXOrigin)] \
	    [expr $innerTop + $_(_labelYOrigin)]
    if { $v(-image) != {} || $v(-bitmap) != {} } {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) 
    } else {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) -justify $_(_just)
    }
    
    $canvas raise $_(_gLabel) $_(_gRegion)
    
    
    set _(_offset)    [expr $innerBottom - $outerTop]
    # height
    set _(_majorDim)  [expr $outerBottom - $outerTop]
    # width
    set _(_minorDim) [expr $right - $_(_left)]
    
    set _(_right)   $right 
    set _(_bottom)  $outerBottom
    
    # draw in correct state...
    if { $_(_selected) } {
	select
    } else {
	deselect
    }
    
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _makeNorthTab
#
# Makes a tab that hangs to the north and opens to the south.
# ----------------------------------------------------------------------
body iwidgets::Tab::_makeNorthTab {canvas} {
    ::variable _tan
    $canvas delete $this
    set _(_gLightOutline) {}
    set _(_gBlackOutline) {}
    
    lappend tagList $this TAB
    
    _createLabel $canvas $tagList
    
    # first get the label width and height
    _calcLabelDim $_(_gLabel)
    
    set bottom [expr $_(_top) + $_(_labelHeight)]
    
    set angleOffset [expr $_(_labelHeight) * $_tan($v(-angle))]
    
    set outerLeft $_(_left)
    set outerRight \
	    [expr $outerLeft + $angleOffset + $_(_labelWidth) + $angleOffset]
    set innerLeft [expr $outerLeft + $angleOffset]
    set innerRight [expr $outerLeft + $angleOffset + $_(_labelWidth)]
    
    # tab area... gets filled either white or selected
    set _(_gRegion) [$canvas create polygon \
	    $outerLeft [expr $bottom + 3]  \
	    $innerLeft [expr $_(_top) + $v(-bevelamount)] \
	    [expr $innerLeft +  $v(-bevelamount)] $_(_top) \
	    [expr $innerRight - $v(-bevelamount)] $_(_top) \
	    $innerRight [expr $_(_top) + $v(-bevelamount)]\
	    $outerRight [expr $bottom + 3] \
	    $outerLeft [expr $bottom + 3] \
	    -tags $tagList  \
	    ]
    
    # lighter shadow (left edge)
    set _(_gLightShadow) [$canvas create line \
	    [expr $outerLeft + 1] [expr $bottom + 3]  \
	    [expr $innerLeft + 1] [expr $_(_top) + $v(-bevelamount)] \
	    [expr $innerLeft + $v(-bevelamount)] [expr $_(_top) + 1]\
	    [expr $innerRight - $v(-bevelamount)] [expr $_(_top) + 1]\
	    -tags $tagList \
	    ]
    
    # darker shadow (bottom and right edges)
    set _(_gDarkShadow) [$canvas create line \
	    [expr $innerRight - $v(-bevelamount)] [expr $_(_top) + 1]\
	    [expr $innerRight - 1] [expr $_(_top) + $v(-bevelamount)]\
	    [expr $outerRight - 1] [expr $bottom + 3]\
	    -tags $tagList \
	    ]
    
    set _(_gLightOutline) [$canvas create line \
	    $outerLeft [expr $bottom + 3]  \
	    $innerLeft [expr $_(_top) + $v(-bevelamount)] \
	    [expr $innerLeft +  $v(-bevelamount)] $_(_top) \
	    [expr $innerRight - $v(-bevelamount)] $_(_top) \
	    -tags $tagList \
	    ]
    
    set _(_gBlackOutline) [$canvas create line \
	    [expr $innerRight - $v(-bevelamount)] $_(_top) \
	    $innerRight [expr $_(_top) + $v(-bevelamount)]\
	    $outerRight [expr $bottom + 3] \
	    $outerLeft [expr $bottom + 3] \
	    -tags $tagList \
	    ]
    
    # top of tab... to make it closed off
    set _(_gTopLine) [$canvas create line \
	    0 0 0 0\
	    -tags $tagList  \
	    ]
    #[expr $outerLeft + 2] [expr $_(_top) + 1] \
	    [expr $outerRight - 2] [expr $_(_top) + 1]
    
    # top of tab... to make it closed off
    set _(_gTopLineShadow) [$canvas create line \
	    0 0 0 0 \
	    -tags $tagList \
	    ]
    #$outerLeft $_(_top) \
	    $outerRight $_(_top) 
    
    $canvas coords $_(_gLabel) [expr $innerLeft + $_(_labelXOrigin)] \
	    [expr $_(_top) + $_(_labelYOrigin)]
    
    if { $v(-image) != {} || $v(-bitmap) != {} } {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) 
    } else {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) -justify $_(_just)
    }
    
    $canvas raise $_(_gLabel) $_(_gRegion)
    
    
    set _(_offset)    [expr $innerRight - $outerLeft]
    # width
    set _(_majorDim)  [expr $outerRight - $outerLeft]
    # height
    set _(_minorDim) [expr $bottom - $_(_top)]
    
    set _(_right)     $outerRight
    set _(_bottom)    $bottom
    
    # draw in correct state...
    if { $_(_selected) } {
	select
    } else {
	deselect
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _makeSouthTab
#
# Makes a tab that hangs to the south and opens to the north.
# ----------------------------------------------------------------------
body iwidgets::Tab::_makeSouthTab {canvas} {
    ::variable _tan
    $canvas delete $this
    set _(_gLightOutline) {}
    set _(_gBlackOutline) {}
    
    lappend tagList $this TAB
    
    _createLabel $canvas $tagList
    
    # first get the label width and height
    _calcLabelDim $_(_gLabel)
    
    set bottom [expr $_(_top) + $_(_labelHeight)]
    
    set angleOffset [expr $_(_labelHeight) * $_tan($v(-angle))]
    
    set outerLeft $_(_left)
    set outerRight \
	    [expr $outerLeft + $angleOffset + $_(_labelWidth) + $angleOffset]
    set innerLeft [expr $outerLeft + $angleOffset]
    set innerRight [expr $outerLeft + $angleOffset + $_(_labelWidth)]
    
    # tab area... gets filled either white or selected
    set _(_gRegion) [$canvas create polygon \
	    $outerLeft [expr $_(_top) + 1] \
	    $innerLeft [expr $bottom  - $v(-bevelamount)]\
	    [expr $innerLeft + $v(-bevelamount)] $bottom \
	    [expr $innerRight - $v(-bevelamount)] $bottom \
	    $innerRight [expr $bottom - $v(-bevelamount)]\
	    $outerRight [expr $_(_top) + 1] \
	    $outerLeft [expr $_(_top) + 1] \
	    -tags $tagList  \
	    ]
    
    
    # lighter shadow (left edge)
    set _(_gLightShadow) [$canvas create line \
	    [expr $outerLeft+1] $_(_top) \
	    [expr $innerLeft+1] [expr $bottom-$v(-bevelamount)] \
	    -tags $tagList \
	    ]
    
    # darker shadow (bottom and right edges)
    set _(_gDarkShadow) [$canvas create line \
	    [expr $innerLeft+1] [expr $bottom-$v(-bevelamount)] \
	    [expr $innerLeft+$v(-bevelamount)] [expr $bottom-1] \
	    [expr $innerRight-$v(-bevelamount)] [expr $bottom-1] \
	    [expr $innerRight-1] [expr $bottom-$v(-bevelamount)] \
	    [expr $outerRight-1] [expr $_(_top) + 1] \
	    -tags $tagList \
	    ]
    # outline of tab
    set _(_gBlackOutline) [$canvas create line \
	    $outerLeft [expr $_(_top) + 1] \
	    $innerLeft [expr $bottom  -$v(-bevelamount)]\
	    [expr $innerLeft + $v(-bevelamount)] $bottom \
	    [expr $innerRight - $v(-bevelamount)] $bottom \
	    $innerRight [expr $bottom - $v(-bevelamount)]\
	    $outerRight [expr $_(_top) + 1] \
	    -tags $tagList \
	    ]
    
    # top of tab... to make it closed off
    set _(_gTopLine) [$canvas create line \
	    $outerLeft [expr $_(_top) + 1] \
	    $outerRight [expr $_(_top) + 1] \
	    -tags $tagList  \
	    ]
    
    # top of tab... to make it closed off
    set _(_gTopLineShadow) [$canvas create line \
	    $outerLeft $_(_top) \
	    $outerRight $_(_top) \
	    -tags $tagList \
	    ]
    
    #$canvas coords $_(_gLabel) [expr $innerLeft + $_(_padX) + 2] \
	    [expr $_(_top) + $_(_padY)]
    $canvas coords $_(_gLabel) [expr $innerLeft + $_(_labelXOrigin)] \
	    [expr $_(_top) + $_(_labelYOrigin)]
    
    if { $v(-image) != {} || $v(-bitmap) != {} } {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) 
    } else {
	$canvas itemconfigure $_(_gLabel) -anchor $v(-anchor) -justify $_(_just)
    }
    $canvas raise $_(_gLabel) $_(_gRegion)
    
    
    set _(_offset)    [expr $innerRight - $outerLeft]
    
    # width
    set _(_majorDim)  [expr $outerRight - $outerLeft]
    
    # height
    set _(_minorDim) [expr $bottom - $_(_top)]
    
    set _(_right)     $outerRight
    set _(_bottom)    $bottom
    
    # draw in correct state...
    if { $_(_selected) } {
	select
    } else {
	deselect
    }
}

# ----------------------------------------------------------------------
# PRIVATE METHOD: _calcLabelDim
#
# Calculate the width and height of the label bbox of labelItem
# can be either text or bitmap (in future also an image)
#
# There are two ways to calculate the label bbox.
#
# First, if the $_(_width) and/or $_(_height) is specified, we will use
# it to determine that dimension(s) width and/or height. For 
# a width/height of 0 we use the labels bbox to
# give us a base width/height.
# Then we add in the padx/pady to determine final bounds.
#
# Uses the following option or option derived variables:
#	-padx     ($_padX - converted to pixels)
#	-pady     ($_padY - converted to pixels)
#	-anchor   ($anchor)
#	-width    ($_(_width)) This is the width for inside tab (label area)
#	-height   ($_(_height)) This is the width for inside tab (label area)
#
# Side Effects:
#	_labelWidth will be set
#	_labelHeight will be set
#	_labelXOrigin will be set
#	_labelYOrigin will be set
# ----------------------------------------------------------------------
body iwidgets::Tab::_calcLabelDim {labelItem} {
    # ... calculate the label width and height
    set labelBBox [$_(_canvas) bbox $labelItem]
    
    if { $_(_width) > 0 } {
	set _(_labelWidth) [expr $_(_width) + ($_(_padX) * 2)]
    } else {
	set _(_labelWidth) [expr \
		([lindex $labelBBox 2] - [lindex $labelBBox 0]) + ($_(_padX) * 2)]
    }
    
    if { $_(_height) > 0 } {
	set _(_labelHeight) [expr $_(_height) + ($_(_padY) * 2)]
    } else {
	set _(_labelHeight) [expr \
		([lindex $labelBBox 3] - [lindex $labelBBox 1]) + ($_(_padY) * 2)]
    }
    
    # ... calculate the label anchor point
    set centerX [expr $_(_labelWidth)/2.0]
    set centerY [expr $_(_labelHeight)/2.0 - 1]
    
    switch $v(-anchor) {
	n {
	    set _(_labelXOrigin) $centerX
	    set _(_labelYOrigin) $_(_padY)
	    set _(_just) center
	}
	s {
	    set _(_labelXOrigin) $centerX
	    set _(_labelYOrigin) [expr $_(_labelHeight) - $_(_padY)]
	    set _(_just) center
	}
	e {
	    set _(_labelXOrigin) [expr $_(_labelWidth) - $_(_padX) - 1]
	    set _(_labelYOrigin) $centerY
	    set _(_just) right
	}
	w {
	    set _(_labelXOrigin) [expr $_(_padX) + 2]
	    set _(_labelYOrigin) $centerY
	    set _(_just) left
	}
	c {
	    set _(_labelXOrigin) $centerX
	    set _(_labelYOrigin) $centerY
	    set _(_just) center
	}
	ne {
	    set _(_labelXOrigin) [expr $_(_labelWidth) - $_(_padX) - 1]
	    set _(_labelYOrigin) $_(_padY)
	    set _(_just) right
	}
	nw {
	    set _(_labelXOrigin) [expr $_(_padX) + 2]
	    set _(_labelYOrigin) $_(_padY)
	    set _(_just) left
	}
	se {
	    set _(_labelXOrigin) [expr $_(_labelWidth) - $_(_padX) - 1]
	    set _(_labelYOrigin) [expr $_(_labelHeight) - $_(_padY)]
	    set _(_just) right
	}
	sw {
	    set _(_labelXOrigin) [expr $_(_padX) + 2]
	    set _(_labelYOrigin) [expr $_(_labelHeight) - $_(_padY)]
	    set _(_just) left
	}
	default {
	    error "bad anchor position: \
		    \"$tabpos\" must be n, ne, nw, s, sw, se, e, w, or center"
	}
    }
}
